import java.util.*;
import java.awt.image.*;

// This class represents the environment that the user will interact with.  It is abstract,
// since more specific information is required for its display.   For example, the
// GridEnvironment is different from the LegoTestPadEnvironment and LabEnvironment.
// The environment in general has a robot that interacts in it as well as a set of user-
// defined traces that display the path of the robot over time.   The path will typically
// represent an estimated trajectory computed via kinematics and odometry.

public abstract class Environment {
	protected Robot						robot;				// The robot operating in the environment
	protected Pose						destination;		// The destination point of the robot in the environment
	protected Vector<Trace>				traces;				// The traces defined by the user
	protected OccupancyGrid				map;				// The computed map displaying all obstacles
	protected BufferedImage 			mapImage;			// Contains the map to be displayed

	public Environment() {
		robot = new GridBot();
System.out.println(robot.getPose());
		traces = new Vector<Trace>();
		map = new OccupancyGrid(getWidthInPixels(), getHeightInPixels());
		mapImage = new BufferedImage(getWidthInPixels(), getHeightInPixels(), BufferedImage.TYPE_INT_ARGB);
		destination = new Pose(0, robot.pose.x/ PIXELS_PER_CM(), robot.pose.y / PIXELS_PER_CM(), 90);
	}

	public abstract void centerRobot();
	public Robot getRobot() { return robot; }
	public Pose getDestination() { return destination; }
	public void setDestination(Pose p) { destination = p; }
	public Vector<Trace> getTraces() { return traces; }
	public OccupancyGrid getMap() { return map; }
	public void setMap(OccupancyGrid newMap) { map = newMap; }
	public BufferedImage getMapImage() { return mapImage; }

	// Return the trace with the given index
	public Trace getTrace(int i) {
		return traces.get(i);
	}

	// Return the trace at the given ID
	public Trace getTraceWithID(int id) {
		for (Trace t: traces)
			if (t.getID() == id) return t;
		return null;
	}

	// Return the trace with the given name
	public Trace getTrace(String n) {
		for (Trace t: traces)
			if (t.getName().equals(n)) return t;
		return null;
	}

	// Erase all traces
	public void eraseTraces() {
		traces.clear();
	}

	// Reset all traces to contain only the given pose
	public void resetTraces(double x, double y, double angle) {
		for (Trace t: traces)
			t.reset(x, y, angle);
	}

	// Create a new trace with the given name and color, then return the trace's unique integer ID
	public int addTrace(String n, int r, int g, int b) {
		//System.out.println("Environment: Adding Trace [" + n + "] ("+r+","+g+","+b+")");
		Trace t = new Trace(n, new java.awt.Color(r, g, b));
		traces.add(t);
		return t.getID();
	}

	// Redraw the map
	public void redrawMap(boolean showMap, boolean showBinary, boolean showEdges) {
		map.drawOnto(mapImage, showMap, showBinary, showEdges);
	}

	public abstract int getWidthInPixels();
	public abstract int getHeightInPixels();
	public abstract double getHeightInCM();
	public abstract void draw(java.awt.Graphics g, boolean solid, boolean outline);
	public static int PIXELS_PER_CM() {return 4;};
}