import java.util.ArrayList;
import java.awt.Polygon;
import java.awt.Point;

// This class is to be written by the user to perform path planning on a vector map
public class LocalizationEstimator {
	public static final int		NUM_SAMPLES = 500;
	public static final double	TOLERANCE = 0.2;

	private ArrayList<Polygon> 	obstacles;
	private ArrayList<Pose>		currentEstimates;
	private int					maxX, minX, maxY, minY;


	// A Path Planner assumes a starting and end location as well as a collection of obstacles
	public LocalizationEstimator(ArrayList<Polygon> ob) {
		currentEstimates = new ArrayList<Pose>();
		setObstacles(ob);
		resetEstimates();
	}

	// These are used to refresh the start or end points or the obstacle collection
	public ArrayList<Pose> getPoseEstimates() { return currentEstimates; }
	public void setObstacles(ArrayList<Polygon> polys) {
		maxX = maxY = -1;
		minX = minY = 9999999;
		obstacles = polys;
		for (Polygon p: obstacles) {
			for (int i=0; i<p.npoints; i++) {
				if (p.xpoints[i] > maxX) maxX = p.xpoints[i];
				if (p.xpoints[i] < minX) minX = p.xpoints[i];
				if (p.ypoints[i] > maxY) maxY = p.ypoints[i];
				if (p.ypoints[i] < minY) minY = p.ypoints[i];
			}
		}
		resetEstimates();
	}


	// Return a pose that does not intersect any obstacles
	private boolean isValidPose(Pose p) {
		// If it is not inside an obstacle, then it is a good one
		for (Polygon q: obstacles) {
			if (q.contains(p.x, p.y))
				return false;
		}
		return true;
	}

	// Return a pose that does not intersect any obstacles
	private Pose getValidPose() {
		while(true) {
			Pose p = new Pose(0.5, Math.random()*(maxX-minX) + minX, Math.random()*(maxY-minY) + minY, Math.random()*360);
			if (isValidPose(p))
				return p;
		}
	}

	// Reset the pose estimates
	public void resetEstimates() {
		currentEstimates.clear();
		int sampleCount = 0;
		while(sampleCount < NUM_SAMPLES*4) {
			Pose p = getValidPose();
			currentEstimates.add(p);
			sampleCount++;
		}
	}

	// Guess the pose of the robot (cm) based on the estimate
	public Pose guessPose() {
		Pose guess = new Pose();
		if (currentEstimates.isEmpty())
			return guess;
		guess.angle = 0;
		for (Pose p: currentEstimates) {
			guess.x = guess.x + p.x;
			guess.y = guess.y + p.y;
			guess.angle = guess.angle + p.angle;
		}
		guess.x = guess.x / currentEstimates.size() / Environment.PIXELS_PER_CM();
		guess.y = guess.y / currentEstimates.size() / Environment.PIXELS_PER_CM();
		guess.angle = guess.angle / currentEstimates.size();

		return guess;
	}
	
	// Return the intersection point of two lines (assumes that they do indeed intersect)
	public Point intersectionPoint(int x1, int y1, int x2, int y2, int x3, int y3, int x4, int y4) {
		int d = (x1-x2)*(y3-y4) - (y1-y2)*(x3-x4);
		if (d == 0)
			return null;

		int xi = ((x3-x4)*(x1*y2-y1*x2)-(x1-x2)*(x3*y4-y3*x4))/d;
		int yi = ((y3-y4)*(x1*y2-y1*x2)-(y1-y2)*(x3*y4-y3*x4))/d;

		Point p = new Point(xi,yi);
		if (xi < Math.min(x1,x2) || xi > Math.max(x1,x2)) return null;
		if (xi < Math.min(x3,x4) || xi > Math.max(x3,x4)) return null;

		return p;
	}


	// Re-estimate the position again based on the given sensor reading (in pixels)
	public void estimateFromReading(double d) {
		
	}

	// Update all the estimates, based on the given forward movement (in pixels)
	public void updateLocation(float distance) {
		
	}

	// Update all the estimates, based on the given change in orientation (in degrees)
	public void updateOrientation(float degrees) {
		
	}
}