import java.io.*;
import java.util.ArrayList;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Point;

public class Node implements Comparable<Node> {
	public static int    RADIUS = 3;

	// These are the instance variables
	private String    		  label;
	private Point     		  location;
	private ArrayList<Edge>   incidentEdges;
	private boolean   		  selected;
	private float			  distance;
	private Node			  previous;

	// Some constructors
	public Node() {
		this("", new Point(0,0));
	}

	public Node(String aLabel) {
		this(aLabel, new Point(0,0));
	}

	public Node(Point aPoint) {
		this("", aPoint);
	}

	public Node(String aLabel, Point aPoint) {
		label = aLabel;
		location = aPoint;
		incidentEdges = new ArrayList<Edge>();
		previous = null;
		distance = -1;
		selected = false;
	}

	// The get & set methods
	public String getLabel() { return label; }
	public Point getLocation() { return location; }
	public float getDistance() { return distance; }
	public Node getPrevious() { return previous; }

	public void setLabel(String newLabel) { label = newLabel; }
	public void setLocation(Point aPoint) { location = aPoint; }
	public void setLocation(int x, int y) { location = new Point(x, y); }
	public void setDistance(float d) { distance = d; }
	public void setPrevious(Node p) { previous = p; }

    public ArrayList<Edge> incidentEdges() { return incidentEdges; }
    public void addIncidentEdge(Edge e) { incidentEdges.add(e); }

	public boolean isSelected() { return selected; }
	public void setSelected(boolean state) { selected = state; }
	public void toggleSelected() { selected = !selected; }

	public ArrayList<Node> neighbours() {
   		ArrayList<Node> result = new ArrayList<Node>();
    	for (Edge e: incidentEdges)
        	result.add(e.otherEndFrom(this));
    	return result;
	}

	// Nodes look like this:  label(12,43)
	public String toString() {
		return(label + "(" + location.x + "," + location.y + ")");
	}

	public void draw(Graphics aPen, int height) {
		if (selected)
			aPen.setColor(Color.red);
		else
			aPen.setColor(Color.yellow);

		// Draw a blue-filled circle around the center of the node
		aPen.fillOval(location.x - RADIUS, height - location.y - RADIUS, RADIUS * 2, RADIUS * 2);

		// Draw a black border around the circle
		aPen.setColor(Color.black);
		aPen.drawOval(location.x - RADIUS, height - location.y - RADIUS, RADIUS * 2, RADIUS * 2);

		// Draw a label at the top right corner of the node
		if (label != "")
			aPen.drawString(label, location.x + RADIUS, height - location.y - RADIUS);
	}

	// Save the node to the given file.  Note that the incident edges are not saved.
	public void saveTo(PrintWriter aFile) {
		aFile.println(label);
		aFile.println(location.x);
		aFile.println(location.y);
		aFile.println(selected);
	}

	// Load a node from the given file.  Note that the incident edges are not connected
	public static Node loadFrom(BufferedReader aFile) throws IOException {
		Node	aNode = new Node();

		aNode.setLabel(aFile.readLine());
		aNode.setLocation(Integer.parseInt(aFile.readLine()),
						  Integer.parseInt(aFile.readLine()));
		aNode.setSelected(Boolean.valueOf(aFile.readLine()).booleanValue());
	    return(aNode);
	}

	public boolean equals(Object obj) {
		if (obj == null) return false;
		if (!(obj instanceof Node)) return false;

		Node n = (Node)obj;
		if (location == null) return false;
		if (n.getLocation() == null) return false;
		return (location.x == n.getLocation().x) && (location.y == n.getLocation().y);

	}

	public int compareTo(Node a) {
		if (!(a instanceof Node)) return 0;
		Node n = (Node)a;

		if (distance < n.distance) return -1;
		if (distance > n.distance) return 1;
		return 0;
	}

}