import java.util.Arrays;
import java.awt.Color;
import java.awt.image.*;
import javax.swing.*;
import java.io.*;
import java.util.ArrayList;
import java.awt.Polygon;
import java.awt.Point;

public class OccupancyGrid {
	public static int 	SCALE = 1;
	public static int	ROUNDNESS = 45;

	protected int 					width;
	protected int 					height;
	protected double				prob[][];		//probability of the grid cells being occupied
	protected byte					binary[][];		//the gird in binary (set by edgeDetector);
	protected ArrayList<Polygon>	polygons;		//the polygons that make up the vector model;
	protected ArrayList<Polygon>	grownPolygons;	//the grown polygons that make up the vector model;
	protected Graph					graph;			//the visibility graph;

	private static final Color c1 = new Color(0, 0, 0, 255);		// unknown
	private static final Color c2 = new Color(255, 220, 0, 255);	// Wall color
	private static final Color c3 = new Color(255, 255, 255, 255);	// Open space color
	private static final Color c4 = new Color(0, 0, 0, 255);		// open area

	private static Color[] greyScalePalette = new Color[256];
	private static Color[] colorPalette = new Color[512];

	// Bounding box
	private int minX = Integer.MAX_VALUE;
	private int minY = Integer.MAX_VALUE;
	private int maxX = Integer.MIN_VALUE;
	private int maxY = Integer.MIN_VALUE;

	private ReadingMapper 			readingMapper;
	private EdgeDetector 			edgeDetector;
	private PathPlanner 			pathPlanner;
	private LocalizationEstimator	localizer;

	public OccupancyGrid(int width, int height) {
		this.width = width;
		this.height = height;
		prob = new double[width][height];
		readingMapper = new ReadingMapper(width, height, prob);
		edgeDetector = new EdgeDetector(width, height, prob);
		pathPlanner = new PathPlanner(new Point(width/2, height/2), new Point(width-1, height/2));
		polygons = new ArrayList<Polygon>();
		grownPolygons = new ArrayList<Polygon>();
		localizer = new LocalizationEstimator(polygons);
		graph = null;	// for now, but will change when computed
	 	binary = null;	// for now, but will change by edge detector

		// Create Color Palette
		for (int i=0; i<256; i++) {
			double ratio = i/255.0;
			double invRatio = 1-ratio;
			int r = (int)(c2.getRed()*ratio   + c1.getRed()*invRatio);
			int g = (int)(c2.getGreen()*ratio + c1.getGreen()*invRatio);
			int b = (int)(c2.getBlue()*ratio  + c1.getBlue()*invRatio);
			int a = (int)(invRatio*255);
			colorPalette[i] = new Color(r,g,b,a);
		}
		for (int i=256; i<512; i++) {
			double ratio = (i-256)/255.0;
			double invRatio = 1-ratio;
			int r = (int)(c3.getRed()*ratio   + c4.getRed()*invRatio);
			int g = (int)(c3.getGreen()*ratio + c4.getGreen()*invRatio);
			int b = (int)(c3.getBlue()*ratio  + c4.getBlue()*invRatio);
			int a = (int)(ratio*255);
			colorPalette[i] = new Color(r,g,b,a);
		}

		// Create Grayscale Palette
		for (int i=0; i<128; i++) {
			double ratio = (127-i)/127.0;
			greyScalePalette[i] = new Color(0,0,0,(int)(ratio*255.0));
		}
		for (int i=128; i<256; i++) {
			double ratio = (i-128)/127.0;
			ratio *= ratio;
			greyScalePalette[i] = new Color(255,255,255,(int)(ratio*255.0));
		}

	}

	public int getWidth() {	return width; }
	public int getHeight() { return height;	}
	public double getCell(int x, int y) { return prob[x][y]; }
	public void incrementCell(int x, int y) { prob[x][y] += 1; }
	public ArrayList<Polygon> getPolygons() { return polygons; }
	public ArrayList<Polygon> getGrownPolygons() { return grownPolygons; }
	public Graph getGraph() { return graph; }
	public PathPlanner getPathPlanner() { return pathPlanner; }
	public LocalizationEstimator getLocalizer() { return localizer; }

	public void setCell(int x, int y, double newValue) {
		if ((x < width) && (x >= 0) && (y < height) && (y >= 0)) {
			prob[x][y] = newValue;

			if (x < minX) minX = x;
			if (y < minY) minY = y;
			if (x > maxX) maxX = x;
			if (y > maxY) maxY = y;
		}
	}

	public int getMaxX() { return maxX;	}
	public int getMaxY() { return maxY; }
	public int getMinX() { return minX; }
	public int getMinY() { return minY; }

	public void clear() {
		for (int x=0; x<width; x++)
			Arrays.fill(prob[x],0.5);
	}

	public double maxValue() {
		double maxValue = 0;

		for (int x=0; x<width; x++)
			for (int y=0; y<height; y++)
				if (prob[x][y] > maxValue)
					maxValue = prob[x][y];
		return maxValue;
	}

	public double minValue() {
		double minValue = 9999;

		for (int x=0; x<width; x++)
			for (int y=0; y<height; y++)
				if (prob[x][y] < minValue)
					minValue = prob[x][y];
		return minValue;
	}

	public void mergeGridIncrementally(OccupancyGrid g) {
		for (int x=0; x<width; x++)
			for (int y=0; y<height; y++)
				if (g.getCell(x,y) > 0) {
					prob[x][y] += g.getCell(x,y);
				}
	}

	// Apply a sensor model distribution to the reading
	public void applyModel(float x, float y, float angle, float d, float angleOffset, float dError, float beamAngle) {
		readingMapper.applyReading(x, y, angle, d, angleOffset, dError, beamAngle);
	}

	// Adjust the binary threshold of the map.  Compute the edges and the polygons
	public void adjustThreshold(float percentage, float errorThreshold) {
		binary = edgeDetector.computeBinary(0.5 + percentage*0.5);
		polygons = edgeDetector.computePolygons(errorThreshold);
		// Flip polygons backwards from CCW to CW, if necessary
		for (Polygon p: polygons) {
			if (!isCW(p))
				reverse(p);
		}
		//pathPlanner.setObstacles(polygons);
	}

	// Compute the Visibility Graph
	public void computeVisibilityGraph(ArrayList<Polygon> obstacles) {
		if (!obstacles.isEmpty())
			graph = pathPlanner.computeVisibilityGraph(obstacles);
	}

    public void drawOnto(BufferedImage image, boolean showMap, boolean showBinary, boolean showEdges) {
//System.out.println("Drawing Map " + showMap);
    	double max = maxValue();
    	double min = minValue();
		double min2 = Math.max(min,0);
		double max2 = Math.min(max,0);
		//System.out.println("MAX: + " + max + ".  MIN: " + min);

    	int[] TempRGBArray = new int[width * height];
    	int index = 0;
    	double val = 0;
		int color = 0;

		// Set the color of each pixel according to the map
    	for (int x=0; x<width; x++) {
    		for (int y=0; y<height; y++) {
				if (prob[x][y] > 0) {
					val = 256 + (((prob[x][y]-min2) / (max-min2)) * 255);
				} else {
					val = ((prob[x][y]-min) / (max2-min)) * 255;
				}
				index = Math.min(511 - (int)val, 511);

				color = colorPalette[index&0x1ff].getRGB();
	    		if (showMap)
	    			TempRGBArray[(height-1-y) * width + x] = color;
	    		else
	    			TempRGBArray[(height-1-y) * width + x] = Color.white.getRGB();
    		}
		}
		image.setRGB(0, 0, width, height, TempRGBArray, 0, width);

		// Now make the binary threshold pixels RED
		if (((showBinary || showEdges) && binary != null)) {
			for (int x=0; x<width; x++) {
	    		for (int y=0; y<height; y++) {
	    			if (showBinary && binary[x][y] == 1)
						image.setRGB(x,height-1-y, Color.RED.getRGB());
	    			else if (showEdges && binary[x][y] == 2)
							image.setRGB(x,height-1-y, Color.BLUE.getRGB());
	    		}
	    	}
		}
	}

	// Save the map to the file with the given name
	public void saveTo(String fileName) {
		try {
			DataOutputStream theFile  = new DataOutputStream(new FileOutputStream(fileName));
			theFile.writeInt(width);
			theFile.writeInt(height);
			for (int x=0; x<width; x++) {
    			for (int y=0; y<height; y++) {
    				theFile.writeDouble(prob[x][y]);
    			}
			}
            theFile.close();
		}
		catch(IOException e) {
			JOptionPane.showMessageDialog(null, "Error saving map file", "Error", JOptionPane.ERROR_MESSAGE);
		}
	}

	// Load the map from the file with the given name
	public void loadFrom(String fileName) {
		if (new File(fileName).exists()) {
			try {
				DataInputStream theFile  = new DataInputStream(new FileInputStream(fileName));
				width = theFile.readInt();
				height = theFile.readInt();
				prob = new double[width][height];
				for (int x=0; x<width; x++) {
    				for (int y=0; y<height; y++) {
    					prob[x][y] = theFile.readDouble();
    				}
				}
				readingMapper = new ReadingMapper(width, height, prob);
				edgeDetector.setCells(prob);
				//pathPlanner.setObstacles(polygons);
            	theFile.close();
			}
			catch(IOException e) {
				JOptionPane.showMessageDialog(null, "Error loading map file", "Error", JOptionPane.ERROR_MESSAGE);
			}
		}
		else
			JOptionPane.showMessageDialog(null, "Error: map file does not exist: " + fileName, "Error", JOptionPane.ERROR_MESSAGE);
	}

	// Save the vector map to the file with the given name
	public void saveVectorMapTo(String fileName) {
		try {
			DataOutputStream theFile  = new DataOutputStream(new FileOutputStream(fileName));
			theFile.writeInt(polygons.size());
			for (Polygon p: polygons) {
				theFile.writeInt(p.npoints);
    			for (int n=0; n<p.npoints; n++) {
    				theFile.writeInt(p.xpoints[n]);
    				theFile.writeInt(p.ypoints[n]);
    			}

			}
            theFile.close();
		}
		catch(IOException e) {
			JOptionPane.showMessageDialog(null, "Error saving map file", "Error", JOptionPane.ERROR_MESSAGE);
		}
	}

	// Load the vector map from the file with the given name
	public void loadVectorMapFrom(String fileName) {
		if (new File(fileName).exists()) {
			try {
				// THIS CODE READS IT IN AS A TEXT FILE
				/*BufferedReader theFile = new BufferedReader(new FileReader(fileName));
				theFile.readLine();
				theFile.readLine();
				int numPolys = Integer.parseInt(theFile.readLine());
				polygons.clear();
				for (int n=0; n<numPolys; n++) {
					Polygon p = new Polygon();
					polygons.add(p);
					theFile.readLine();
					int size = Integer.parseInt(theFile.readLine());
					for (int i=0; i<size; i++) {
						java.util.StringTokenizer st = new java.util.StringTokenizer(theFile.readLine(), "(,)");
						p.addPoint((int)(Integer.parseInt(st.nextToken()) * 1.1)+20, (int)(Integer.parseInt(st.nextToken()) * 1.1)+20);
					}
					theFile.readLine();
				}*/

				DataInputStream theFile  = new DataInputStream(new FileInputStream(fileName));
				int numPolys = theFile.readInt();
				polygons.clear();
				for (int n=0; n<numPolys; n++) {
					Polygon p = new Polygon();
					polygons.add(p);
					int size = theFile.readInt();
					for (int i=0; i<size; i++)
						p.addPoint(theFile.readInt(), theFile.readInt());
				}

				/*polygons.clear();
				Polygon p = new Polygon();
				p.addPoint(200, 150);
				p.addPoint(200, 514);
				p.addPoint(211, 514);
				p.addPoint(211, 150);
				polygons.add(p);

				p = new Polygon();
				p.addPoint(211, 514);
				p.addPoint(575, 514);
				p.addPoint(575, 503);
				p.addPoint(211, 503);
				polygons.add(p);

				p = new Polygon();
				p.addPoint(575, 514);
				p.addPoint(586, 514);
				p.addPoint(586, 393);
				p.addPoint(575, 393);
				polygons.add(p);

				p = new Polygon();
				p.addPoint(454, 404);
				p.addPoint(575, 404);
				p.addPoint(575, 393);
				p.addPoint(454, 393);
				polygons.add(p);

				p = new Polygon();
				p.addPoint(575, 393);
				p.addPoint(586, 393);
				p.addPoint(586, 272);
				p.addPoint(575, 272);
				polygons.add(p);

				p = new Polygon();
				p.addPoint(464, 272);
				p.addPoint(585, 272);
				p.addPoint(585, 261);
				p.addPoint(464, 261);
				polygons.add(p);

				p = new Polygon();
				p.addPoint(453, 272);
				p.addPoint(464, 272);
				p.addPoint(464, 150);
				p.addPoint(453, 150);
				polygons.add(p);

				p = new Polygon();
				p.addPoint(332, 161);
				p.addPoint(453, 161);
				p.addPoint(453, 150);
				p.addPoint(332, 150);
				polygons.add(p);

				p = new Polygon();
				p.addPoint(211, 161);
				p.addPoint(332, 161);
				p.addPoint(332, 150);
				p.addPoint(211, 150);
				polygons.add(p);*/

				//pathPlanner.setObstacles(polygons);
				localizer.setObstacles(polygons);
				computeVisibilityGraph(polygons);
            	theFile.close();
			}
			catch(IOException e) {
				JOptionPane.showMessageDialog(null, "Error loading map file", "Error", JOptionPane.ERROR_MESSAGE);
			}
		}
		else
			JOptionPane.showMessageDialog(null, "Error: map file does not exist: " + fileName, "Error", JOptionPane.ERROR_MESSAGE);
	}


	// Create a new map, based on the polygons
	public ArrayList<Polygon> createFrom(ArrayList<Polygon> originalPolys, ArrayList<Polygon> expandedPolys, BufferedImage mapImage) {
		// Create a temporary image on which to draw the polygons
		BufferedImage tempImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
		tempImage.getGraphics().setColor(Color.GREEN);
		for (Polygon p: expandedPolys) {
			tempImage.getGraphics().fillPolygon(p);
		}

		// Now extract the probabilities (1 for polygon, 0 otherwise);
		//int polyColor = EnvironmentPanel.POLYGON_FILL_COLOR.getRGB();
		for (int x=0; x<width; x++) {
    		for (int y=0; y<height; y++) {
    			if (tempImage.getRGB(x,y) == -1)
    				prob[x][y] = 1;
    			else
    				prob[x][y] = 0;
    		}
		}
		edgeDetector.setCells(prob);
		binary = edgeDetector.computeBinary(0.5f);
		ArrayList<Polygon> newPolys = edgeDetector.computePolygons(1);

		// Flip polygons backwards from CCW to CW, if necessary
		for (Polygon p: newPolys) {
			if (!isCW(p))
				reverse(p);
		}

		// Remove any polygons that are actually holes
		java.util.Iterator<Polygon>	iterator = newPolys.iterator();
		while(iterator.hasNext()) {
			Polygon p = iterator.next();
			// Check all original polygons to see if any lie within here, otherwise it is a hole
			boolean hole = true;
			for (Polygon obst: originalPolys) {
				for (int i=0; i<obst.npoints; i++) {
					if (p.contains(obst.xpoints[i], obst.ypoints[i]))
						hole = false;
				}
			}
			if (hole)
				iterator.remove();
		}

		//pathPlanner.setObstacles(newPolys);

		return newPolys;
	}


	// Return whether or not the given polygon is CW as oposed to CCW oriented
	private boolean isCW(Polygon p) {
		int total = 0;
		for (int i=0; i<p.npoints; i++) {
			total += (p.xpoints[(i+1)%p.npoints] - p.xpoints[i]) * (p.ypoints[(i+1)%p.npoints] + p.ypoints[i]);
		}
		return total > 0;
	}

	// Reverse a polygon's points
	private void reverse(Polygon p) {
		int[] xpoints = new int[p.npoints];
		int[] ypoints = new int[p.npoints];
		for (int i=p.npoints-1; i>=0; i--) {
			xpoints[i] = p.xpoints[p.npoints-1-i];
			ypoints[i] = p.ypoints[p.npoints-1-i];
		}
		for (int i=0; i<p.npoints; i++) {
			p.xpoints[i] = xpoints[i];
			p.ypoints[i] = ypoints[i];
		}
	}

	// Return the distance from the robot to the closest polygon (if there is a polygon)
	public int simulateReading(double x, double y, double angle) {
		if (polygons.isEmpty())
			return -1;

		int x1 = (int)x;
		int y1 = (int)y;
		int x2 = (int)(x + 1000*Math.cos(Math.toRadians(angle)));
		int y2 = (int)(y + 1000*Math.sin(Math.toRadians(angle)));

		int intersectionDistance = 1000;
		for (Polygon q: polygons) {
			for (int i=0; i<q.npoints; i++) {
				// Determine if there is an obstacle boundary between the robot and the distance
				// reading.  If so, keep the closest intersection point
				if (java.awt.geom.Line2D.Double.linesIntersect(x1, y1, x2, y2, q.xpoints[i], q.ypoints[i], q.xpoints[(i+1)%q.npoints], q.ypoints[(i+1)%q.npoints])) {
					Point intersection = intersectionPoint(x1, y1, x2, y2, q.xpoints[i], q.ypoints[i], q.xpoints[(i+1)%q.npoints], q.ypoints[(i+1)%q.npoints]);
					if (intersection != null) {
						int dist = (int)intersection.distance(x, y);
						if (dist < intersectionDistance) {
							intersectionDistance = dist;
						}
					}
				}
			}
		}

		if (intersectionDistance == 10000) return -1;

////System.out.println("(x,y,a) = " + x+","+y+","+angle+") Intersects at distance: " + intersectionDistance);

		return intersectionDistance;
	}

	// Return the intersection point of two lines (assumes that they do indeed intersect)
	public Point intersectionPoint(int x1,int y1,int x2,int y2, int x3, int y3, int x4,int y4) {
		int d = (x1-x2)*(y3-y4) - (y1-y2)*(x3-x4);
		if (d == 0)
			return null;

		int xi = ((x3-x4)*(x1*y2-y1*x2)-(x1-x2)*(x3*y4-y3*x4))/d;
		int yi = ((y3-y4)*(x1*y2-y1*x2)-(y1-y2)*(x3*y4-y3*x4))/d;

		Point p = new Point(xi,yi);
		if (xi < Math.min(x1,x2) || xi > Math.max(x1,x2)) return null;
		if (xi < Math.min(x3,x4) || xi > Math.max(x3,x4)) return null;

		return p;
	}
}