// This class handles all bluetooth/USB communications between the RTMS and the robot.
// It is a process that runs continually so that information can be sent and
// received at any time without interfering with the main user interface

import java.io.*;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;

public class PCCommunicator extends Thread {
	// Communication-specific variables and constants
	public static final int 	SERVER_PORT = 6000; 		// arbitrary, but above 1023
	public static final int 	CONNECTION_TIMEOUT = 2000; 	// maximum time to wait (ms) for a connection to EV3 before giving up
	public static final String	EV3_ADDRESS = "10.0.1.1"; 	// IP address of the EV3

	// These constants define the possible communication status between the robot and pc at any time
	public static final byte 	NOT_CONNECTED = 0;
	public static final byte 	CONNECTED = 1;
	public static final byte 	READY_TO_START = 2;
	public static final byte 	START_PRESSED = 3;
	public static final byte 	RUNNING = 4;

	// These constants represent incoming commands from the robot
	public static final byte	CONNECT_ME = 0;				// Tell the RTMS that the robot is up-and-running ...
	public static final byte	DEFINE_NEW_TRACE = 1;		// Ask the RTMS to make a new trace
	public static final byte	REQUEST_START = 2;			// Request the start signal from the RTMS
	public static final byte	GET_POSE = 3;				// Request the current pose from the RTMS
	public static final byte	SEND_POSE = 4;				// Send a pose to the RTMS, to be attached to a trace
	public static final byte	REQUEST_RUN_STATUS = 5;		// Request the run status of the trace
	public static final byte	RESET_TRACE = 6;			// Clear the current trace
	public static final byte	DEFINE_OBSTACLE_SET = 7;	// Ask the RTMS to make a new obstacle set
	public static final byte	SEND_OBSTACLE = 8;			// Send an obstacle estimate to the RTMS, to be attached to an obstacle set
	public static final byte	SEND_READING = 9;			// Send a sensor distance reading to the RTMS, to be mapped
	public static final byte	RESET_MAP = 10;				// Clear the current map
	public static final byte	DEBUG_PRINT = 11;			// Request from robot to print a string for debugging
	public static final byte	MOVED_FORWARD = 12;			// Tell the RTMS that the robot moved forward a certain amount (for localization)
	public static final byte	TURNED = 13;				// Tell the RTMS that the robot turned a certain amount (for localization)
	public static final byte	SEE_AHEAD = 14;				// Tell the RTMS that the robot sees an object at a certain distance ahead (for localization)
	public static final byte	SHUT_DOWN = 15;				// Tell the RTMS that the robot is all done now and shutting down
	
	protected boolean			requestStop;				// This is set to true when the user presses the STOP button
	protected boolean			requestShutdown;			// This is set to true when the communicator needs to be shut down

	protected RTMS				rtms;						// The main program to communicate with
	protected int				connectionStatus;			// The current connection status of the robot
	protected Socket			robotSocket;				// The TCP socket that connects to the robot
	protected DataOutputStream	robotOutputStream;			// This is the stream to send data to the robot
	protected DataInputStream	robotInputStream;			// This is the stream to receive data from the robot


    public PCCommunicator(RTMS mainProgram) {
    	rtms = mainProgram;
    	setConnectionStatus(NOT_CONNECTED);
    	requestStop = false;
    }

	public RTMS getRTMS() { return rtms; }
	public int getConnectionStatus() { return connectionStatus; }
	public void setConnectionStatus(byte status) {
		connectionStatus = status;
		System.out.println("Connection Status: " + connectionStatus);
	}


	// This is the main method for communicating with the robot in an infinite loop.
	public void run() {
		byte 	command;	// an incoming command from the robot
		int		id = 0;		// a Trace id

		// Loop unless the RTMS wants to shut down
		while(true) {
			try {
				switch (connectionStatus) {
					case NOT_CONNECTED:
						// Nothing to do if not connected... this should never happen at this point
						break;
					case CONNECTED:
						// Wait for an incoming command from the robot
						command = robotInputStream.readByte();
						switch(command) {
							case DEFINE_NEW_TRACE: 	// an incoming request to define a new trace
								System.out.println("PCcom Received Define Trace Command");
								int r = (int)robotInputStream.readInt();	// Get the red component of the trace color
								int g = (int)robotInputStream.readInt();	// Get the green component of the trace color
								int b = (int)robotInputStream.readInt();	// Get the blue component of the trace color
								String name = robotInputStream.readUTF();	// Get the trace name
								id = rtms.defineTrace(name, r, g, b);		// Create the trace
								System.out.println("PCcom [" + name + "] ("+r+","+g+","+b+")");
								robotOutputStream.writeInt(id);			// Send back the trace ID
								robotOutputStream.flush();
								break;
							case DEFINE_OBSTACLE_SET: 	// an incoming request to define a new obstacle set
								System.out.println("PCcom Received Define Obstacle Set Command");
								r = (int)robotInputStream.readInt();	// Get the red component of the obstacle set color
								g = (int)robotInputStream.readInt();	// Get the green component of the obstacle set color
								b = (int)robotInputStream.readInt();	// Get the blue component of the obstacle set color
								name = robotInputStream.readUTF();		// Get the obstacle set name
								int index = rtms.createObstacleSet(name, r, g, b);		// Create the obstacle set
								System.out.println("PCcom [" + name + "] ("+r+","+g+","+b+")");
								robotOutputStream.writeInt(index);		// Send back the index
								robotOutputStream.flush();
								break;
							case REQUEST_START: 	// an incoming request for robot to start
								System.out.println("PCcom Received Request Start Command");
								setConnectionStatus(READY_TO_START);
								rtms.update();	// Inform the GUI
								break;
							default:
								System.out.println("PCcom Received Unknown Command: " + command);
								break;
						}
						break;
					case READY_TO_START:
						// Do Nothing
						break;
					case START_PRESSED:
						System.out.println("PCcom Sending Start ACK to Robot");
						// The start button was just pressed, so tell the robot to begin
						robotOutputStream.writeByte(255);		// Send back the Start signal (value of 255)
						robotOutputStream.flush();
						setConnectionStatus(RUNNING);
						rtms.update();	// Inform the GUI
						break;
					case RUNNING:
						// Wait for an incoming command from the robot
						command = robotInputStream.readByte();
						switch(command) {
							case GET_POSE: 	// an incoming request for the current pose
								//System.out.println("RTMS Received Get Pose Command");
								Pose p = rtms.poseRequestObtained();
								robotOutputStream.writeFloat((float)p.x);		// Send back the pose
								robotOutputStream.writeFloat((float)p.y);
								robotOutputStream.writeFloat((float)p.angle);
								robotOutputStream.flush();
								break;
							case SEND_POSE: 	// an incoming set of pose data to follow
								//System.out.println("RTMS Received Pose From Robot");
								id = robotInputStream.readInt();
								rtms.poseObtained(id, robotInputStream.readFloat(), robotInputStream.readFloat(), robotInputStream.readFloat());
								robotOutputStream.writeBoolean(!requestStop);
								robotOutputStream.flush();
								rtms.update();	// Inform the GUI
								break;
							case SEND_OBSTACLE: 	// an incoming obstacle position estimate
								//System.out.println("RTMS Received Obstacle From Robot");
								int index = robotInputStream.readInt();	// Get the index of this obstacle set within the trace
								rtms.obstacleObtained(index, robotInputStream.readFloat(), robotInputStream.readFloat());
								robotOutputStream.writeBoolean(!requestStop);
								robotOutputStream.flush();
								rtms.update();	// Inform the GUI
								break;
							case SEND_READING: 	// an incoming distance reading to be mapped
								//System.out.println("RTMS Received Obstacle From Robot");
								index = robotInputStream.readInt();	// Get the index of this obstacle set within the trace
								rtms.readingObtained(index, robotInputStream.readFloat(), robotInputStream.readFloat(),
													 robotInputStream.readFloat(), robotInputStream.readFloat(), robotInputStream.readFloat(),
													 robotInputStream.readFloat(), robotInputStream.readFloat());
								robotOutputStream.writeBoolean(!requestStop);
								robotOutputStream.flush();
								rtms.update();	// Inform the GUI
								break;
							case MOVED_FORWARD: 	// an incoming update on the robot's forward change in position
								//System.out.println("RTMS Received Moved Forward Update From Robot");
								rtms.moveForwardObtained(robotInputStream.readFloat());
								robotOutputStream.writeBoolean(!requestStop);
								robotOutputStream.flush();
								rtms.update();	// Inform the GUI
								break;
							case TURNED: 			// an incoming update on the robot's orientation change
								//System.out.println("RTMS Received Turned Update From Robot");
								rtms.turnObtained(robotInputStream.readFloat());
								robotOutputStream.writeBoolean(!requestStop);
								robotOutputStream.flush();
								rtms.update();	// Inform the GUI
								break;
							case SEE_AHEAD: // an incoming distance reading ahead from the robot
								//System.out.println("RTMS Received See Ahead Update From Robot");
								rtms.distanceObtained(robotInputStream.readFloat()*Environment.PIXELS_PER_CM());
								robotOutputStream.writeBoolean(!requestStop);
								robotOutputStream.flush();
								rtms.update();	// Inform the GUI
								break;
							case SHUT_DOWN: 	// an incoming request to shut down
								setConnectionStatus(NOT_CONNECTED);
								rtms.update();	// Inform the GUI
								break;
							case RESET_TRACE: 	// an incoming request to erase the current trace data and start over
								//System.out.println("RTMS Received Reset Trace Command");
								id = robotInputStream.readInt();
								rtms.resetTrace(id, robotInputStream.readFloat(), robotInputStream.readFloat(), robotInputStream.readFloat());
								robotOutputStream.writeBoolean(!requestStop);
								robotOutputStream.flush();
								rtms.update();	// Inform the GUI
								break;
							case RESET_MAP: 	// an incoming request to erase the current map data and start over
								//System.out.println("RTMS Received Reset Map Command");
								id = robotInputStream.readInt();
								robotOutputStream.writeBoolean(!requestStop);
								robotOutputStream.flush();
								rtms.resetMap();
								rtms.update();	// Inform the GUI
								break;
							case DEBUG_PRINT: 	// an incoming request to display a string for debugging
								//System.out.println("RTMS Received Debug Print Command");
								rtms.debugPrint(robotInputStream.readUTF());
								robotOutputStream.writeBoolean(!requestStop);
								robotOutputStream.flush();
								break;
							default:
								//System.out.println("RTMS Received Unknown Command: " + command);
								break;
						}
				}
			} catch (IOException ex) {
				setConnectionStatus(NOT_CONNECTED);
				rtms.update();	// Inform the GUI
			}
		}
	}

	// This method resets the server socket
	public boolean establishEV3Connection() {
		requestStop = false;
		
		// Try connecting to the EV3
		setConnectionStatus(NOT_CONNECTED);
		try {
			System.out.println("Tring to connect to EV3 ...");
			robotSocket = new Socket();
			robotSocket.connect(new InetSocketAddress(InetAddress.getByName(EV3_ADDRESS), SERVER_PORT), CONNECTION_TIMEOUT);
			//robotSocket = new Socket(InetAddress.getByName(EV3_ADDRESS), SERVER_PORT);
			System.out.println("Connected to EV3.");
		}
		catch (IOException e) {
			robotSocket = null;
			System.out.println("ERROR connecting to EV3");
			return false;
		}
		try {
			robotInputStream = new DataInputStream(robotSocket.getInputStream());
			robotOutputStream = new DataOutputStream(robotSocket.getOutputStream());
		}
		catch (IOException e) {
			System.out.println("ERROR streaming to EV3");
			return false;
		}
		setConnectionStatus(CONNECTED);
		System.out.println("Connected to EV3.");
		
		return true;
	}
		

	// This method closes the robot socket
	private void closeServerSocket() {
		try {
			if (robotOutputStream != null)
				robotOutputStream.close();
			if (robotInputStream != null)
				robotInputStream.close();
			if (robotSocket != null)
				robotSocket.close();
		}
		catch (IOException e) {
			System.out.println("ERROR closing socket connection to robot");
		}
		setConnectionStatus(NOT_CONNECTED);
	}

	
	// Called when the START button is pressed
	public void startTrace() {
		setConnectionStatus(START_PRESSED);
	}

	// Called when the STOP button is pressed
	public void stopTrace() {
		requestStop = true;
	}

	// Called when the communicator needs to be shut down (e.g., a different communicator will be used)
	public void shutdown() {
		closeServerSocket();
		requestShutdown = true;
	}
}