import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.filechooser.FileNameExtensionFilter;


// This is the main program for the Robot Tracking and Mapping System.
// It represents the main window and also the controller for all user-interface
// interaction.

public class RTMS extends JFrame {
	public static final long serialVersionUID = 1234567890;
	
	private Environment				model;				// The environment model in which the robot is running
	private RTMSView				view;				// The view for the GUI
	public  DebugDialog 			debugDialog;

	private PCCommunicator			pcCommunicator;		// This handles all communication between the robot and the PC

	// Controller-specific variables
	private boolean					moveRobot;			// TRUE if the robot is being moved
	private boolean					spinRobot;			// TRUE if the robot is being spun
	private boolean					moveDestination;	// TRUE if the robot's destination point is being moved
	private boolean					spinDestination;	// TRUE if the robot's destination orientation is being spun
	private int						lastMouseX;			// Last x value of mouse

	// This is the constructor
	public RTMS() {
		super("Robot Tracking and Mapping System");

		// Create the debug window
		debugDialog = new DebugDialog(this);

		// Create and start a communication process for communicating with the PC
		pcCommunicator = new PCCommunicator(this);
		pcCommunicator.start();

		// Create the model, which is the environment
		model = new Lab5174Environment();
	
		// Create and add the view
		getContentPane().add(view = new RTMSView(this, model, pcCommunicator));
		setJMenuBar(view.getMenuBar());

		// Now add the listeners
		view.getResetPoseButton().addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				model.centerRobot();
				update();
			}});
		
		view.getConnectButton().addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				handleConnectButton();
			}});

		// Handle the Show Robot button
		view.getShowRobotButton().addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				update();
			}});

		// Handle the robot scale slider
		view.getScaleSlider().addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					float value = view.getScaleSlider().getValue()/100.0f;
					if (value > 0)
						Robot.SCALE = 1.25f + value/2;
					else
						Robot.SCALE = 1.25f + value/2;
					update();
				}});

		// Handle a clear traces press
		view.getClearTracesButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					for(Trace t: model.getTraces())
						t.getPoses().clear();
					model.resetTraces(model.getRobot().getPose().x, model.getRobot().getPose().y, model.getRobot().getPose().angle);
					model.getMap().clear();
					
					update();
				}});

		// Handle a clear maps press
		view.getClearMapsButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (model.getMap() != null) {
						model.getMap().clear();
						update();
					}
				}});

		// Handle a save Grid map press
		view.getSaveGridMapButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					JFileChooser chooser = new JFileChooser(new File("."));
					FileNameExtensionFilter filter = new FileNameExtensionFilter("Map File", "map");
    				chooser.setFileFilter(filter);
					int returnVal = chooser.showSaveDialog(null);
					if (returnVal == JFileChooser.APPROVE_OPTION) {
						String n = chooser.getSelectedFile().getName();
						if (!n.toUpperCase().endsWith(".MAP"))
							n += ".map";
						model.getMap().saveTo(n);
					}
				}});

		// Handle a load Grid map press
		view.getLoadGridMapButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					JFileChooser chooser = new JFileChooser(new File("."));
					FileNameExtensionFilter filter = new FileNameExtensionFilter("Map File", "map");
    				chooser.setFileFilter(filter);
					int returnVal = chooser.showOpenDialog(null);
					if (returnVal == JFileChooser.APPROVE_OPTION) {
						// First disable all PathPlanning buttons
						view.getShowDestinationButton().setSelected(false);
						view.getShowSupportsButton().setSelected(false);
						view.getShowVisibilityButton().setSelected(false);
						view.getShowPathButton().setSelected(false);
						view.getShowGrownButton().setSelected(false);


						model.getMap().loadFrom(chooser.getSelectedFile().getName());
						model.getMap().adjustThreshold(view.getThresholdSlider().getValue()/100.0f, view.getErrorThresholdSlider().getValue());
					}
					update();
				}});

		// Handle a save Vector map press
		view.getSaveVectorMapButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					JFileChooser chooser = new JFileChooser(new File("."));
					FileNameExtensionFilter filter = new FileNameExtensionFilter("Vector Map File", "vmp");
    				chooser.setFileFilter(filter);
					int returnVal = chooser.showSaveDialog(null);
					if (returnVal == JFileChooser.APPROVE_OPTION) {
						String n = chooser.getSelectedFile().getName();
						if (!n.toUpperCase().endsWith(".VMP"))
							n += ".vmp";
						model.getMap().saveVectorMapTo(n);
					}
				}});

		// Handle a load Vector map press
		view.getLoadVectorMapButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					JFileChooser chooser = new JFileChooser(new File("."));
					FileNameExtensionFilter filter = new FileNameExtensionFilter("Vector Map File", "vmp");
    				chooser.setFileFilter(filter);
					int returnVal = chooser.showOpenDialog(null);
					if (returnVal == JFileChooser.APPROVE_OPTION) {
						model.getMap().loadVectorMapFrom(chooser.getSelectedFile().getName());
						view.getShowPolygonsButton().setSelected(true);
						update();
					}
				}});

		// Handle the show/hide map button
		view.getShowMapButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					update();
				}});

		// Handle the Binary Map Threshold slider
		view.getThresholdSlider().addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					model.getMap().adjustThreshold((int)view.getThresholdSlider().getValue()/100.0f, view.getErrorThresholdSlider().getValue()/2.0f);
					update();
				}});

		// Handle the Polygon Error Threshold slider
		view.getErrorThresholdSlider().addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					model.getMap().adjustThreshold((int)view.getThresholdSlider().getValue()/100.0f, view.getErrorThresholdSlider().getValue()/2.0f);
					update();
				}});

		// Handle the show/hide binary button
		view.getShowBinaryButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					update();
				}});
		// Handle the show/hide edges button
		view.getShowEdgesButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					update();
				}});

		// Handle the show/hide polygons button
		view.getShowPolygonsButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					update();
				}});

		// Handle the show/hide destination button
		view.getShowDestinationButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					update();
				}});

		// Handle the show/hide support lines button
		view.getShowSupportsButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					update();
				}});

		// Handle the show/hide visibiliy graph button
		view.getShowVisibilityButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					update();
				}});

		// Handle the show/hide shortest path button
		view.getShowPathButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					view.getShowTurnPathButton().setSelected(false);
					update();
				}});

		// Handle the show/hide shortest path button
		view.getShowTurnPathButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					view.getShowPathButton().setSelected(false);
					update();
				}});

		// Handle the show/hide Grown Obstacle Space button
		view.getShowGrownButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					update();
				}});

		// Handle the roundness slider
		view.getRoundnessSlider().addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					OccupancyGrid.ROUNDNESS = view.getRoundnessSlider().getValue();
					update();
				}});

		// Handle the show/hide localization button
		view.getShowLocalizationButton().addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					update();
				}});

		/*// Handle the distance reading for localization
		view.getTempButton1().addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				//for (int i=0; i<25; i++)
					distanceObtained(model.getMap().simulateReading(model.getRobot().getPose().x*Environment.PIXELS_PER_CM(), model.getRobot().getPose().y*Environment.PIXELS_PER_CM(), model.getRobot().getPose().angle));
			}});
		// Handle the moving forward for localization
		view.getTempButton2().addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				String input = JOptionPane.showInputDialog(null, "Enter Forward Distance (cm)");
				if (input != null) {
					float d = Float.parseFloat(input);
					model.getRobot().pose.x = model.getRobot().pose.x + d*Math.cos(Math.toRadians(model.getRobot().pose.angle));
					model.getRobot().pose.y = model.getRobot().pose.y + d*Math.sin(Math.toRadians(model.getRobot().pose.angle));
					moveForwardObtained(d);
				}
			}});
		// Handle the turning for localization
		view.getTempButton3().addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				String input = JOptionPane.showInputDialog(null, "Enter Turn Amount (deg)");
				if (input != null) {
					float d = Float.parseFloat(input);
					model.getRobot().pose.angle = model.getRobot().pose.angle + d;
					turnObtained(d);
				}
			}});*/

		// Listen for a mouse press in the environment panel ... remember which button was pressed.
		view.getEnvironmentPanel().addMouseListener(new MouseAdapter() {
			public void mousePressed(MouseEvent e) {
				// Check to see if the robot is being moved or spun
				moveRobot = false;
				if ((e.getButton() == MouseEvent.BUTTON1) && (e.getPoint().distance(model.getRobot().pose.x*(double)Environment.PIXELS_PER_CM(), model.getHeightInPixels() - model.getRobot().pose.y*(double)Environment.PIXELS_PER_CM())<Robot.GRAB_RADIUS*Robot.SCALE)) {
					moveRobot = true;
					return;
				}
				spinRobot = ((e.getButton() != MouseEvent.BUTTON1) && (e.getPoint().distance(model.getRobot().pose.x*(double)Environment.PIXELS_PER_CM(), model.getHeightInPixels() - model.getRobot().pose.y*(double)Environment.PIXELS_PER_CM())<Robot.GRAB_RADIUS*Robot.SCALE));

				// Check to see if the robot's destination is being moved or spun
				moveDestination = ((e.getButton() == MouseEvent.BUTTON1) && (e.getPoint().distance(model.getDestination().x*(double)Environment.PIXELS_PER_CM(), model.getHeightInPixels() - model.getDestination().y*(double)Environment.PIXELS_PER_CM())<Robot.GRAB_RADIUS*Robot.SCALE));
				spinDestination = ((e.getButton() != MouseEvent.BUTTON1) && (e.getPoint().distance(model.getDestination().x*(double)Environment.PIXELS_PER_CM(), model.getHeightInPixels() - model.getDestination().y*(double)Environment.PIXELS_PER_CM())<Robot.GRAB_RADIUS*Robot.SCALE));
			}});

		// Listen for mouse movement.  If left click movement, move robot to mouse position.
		// If right click movement, spin the robot left/right according to the mouse's x position.
		view.getEnvironmentPanel().addMouseMotionListener(new MouseMotionAdapter() {
			public void mouseDragged(MouseEvent e) {
				// Change the robot's pose if necessary
				if (moveRobot) {
					model.getRobot().pose.x = e.getX() / (double)Environment.PIXELS_PER_CM();
					model.getRobot().pose.y = (model.getHeightInPixels() - e.getY()) / (double)Environment.PIXELS_PER_CM();
				}
				else if (spinRobot) {
					if (e.getX() < lastMouseX) {
						model.getRobot().pose.angle = ((int)model.getRobot().pose.angle)/15*15 + 15;
						if (model.getRobot().pose.angle > 180)
							model.getRobot().pose.angle -= 360;
					}
					if (e.getX() > lastMouseX) {
						model.getRobot().pose.angle = ((int)model.getRobot().pose.angle)/15*15 - 15;
						if (model.getRobot().pose.angle < -180)
							model.getRobot().pose.angle += 360;
					}
				}

				// Change the robot's destination if necessary
				else if (moveDestination) {
					model.getDestination().x = e.getX() / (double)Environment.PIXELS_PER_CM();
					model.getDestination().y = (model.getHeightInPixels() - e.getY()) / (double)Environment.PIXELS_PER_CM();
				}
				else if (spinDestination) {
					if (e.getX() < lastMouseX) {
						model.getDestination().angle = ((int)model.getDestination().angle)/15*15 + 15;
						if (model.getDestination().angle > 180)
							model.getDestination().angle -= 360;
					}
					if (e.getX() > lastMouseX) {
						model.getDestination().angle = ((int)model.getDestination().angle)/15*15 - 15;
						if (model.getDestination().angle < -180)
							model.getDestination().angle += 360;
					}
				}
				lastMouseX = e.getX();

				update();
			}});

		setDefaultCloseOperation(EXIT_ON_CLOSE);
		setSize(model.getWidthInPixels()+7, model.getHeightInPixels()+ 72);
		setResizable(false);
		update();
	}

	// Handle the selecting of an environment
	/*private void handleEnvironmentChange(int type) {
		switch(type) {
			case 0: model = new Lab5174Environment(); break;
			case 1: model = new LegoTestPadEnvironment(); break;
			case 2: model = new GridEnvironment(); break;
		};
		view.getEnvironmentPanel().setEnvironment(model);
		view.getEnvironmentPanel().resizeToEnvironment();
		setSize(model.getWidthInPixels()+7, model.getHeightInPixels()+ 72);
		update();
	};*/

	// Handle the pressing of the "Connect/Start/Stop" button
	private void handleConnectButton() {
		switch(pcCommunicator.getConnectionStatus()) {
			case PCCommunicator.NOT_CONNECTED:
				pcCommunicator.establishEV3Connection();
				eraseTraces();// Clear everything to get ready for another round
				update();
				break;
			case PCCommunicator.CONNECTED:
				update();
				break;
			case PCCommunicator.READY_TO_START:
				model.resetTraces(model.getRobot().getPose().x, model.getRobot().getPose().y, model.getRobot().getPose().angle);
				model.getMap().clear();
				update();
				pcCommunicator.startTrace();
				break;
			case PCCommunicator.RUNNING:
				pcCommunicator.stopTrace();
				update();
		}
	}

	// Called when the connection is restarted in order to erase all traces and get ready for the next round
	public void eraseTraces() {
		model.eraseTraces();
		update();
	}
	
	// Called from from PCCommunicator to define a new trace and return its id
	public int defineTrace(String name, int r, int g, int b) {
		int id = model.addTrace(name, r, g, b);
		view.updateTraceMenu();
		model.resetTraces(model.getRobot().getPose().x, model.getRobot().getPose().y, model.getRobot().getPose().angle);
		model.getMap().clear();
		update();
		return id;
	}

	// Called from from PCCommunicator to create a new Obstacle Set and return its index in the trace's list
	public int createObstacleSet(String name, int r, int g, int b) {
		Trace t = model.getTraces().get(0);
		ObstacleSet os = new ObstacleSet(name, r, g, b);
		t.getObstacles().add(os);
		view.updateMapMenu();
		update();
		return t.getObstacles().indexOf(os);
	}

	// Called from PCCommunicator when the robot requests a pose (e.g., starting pose)
	public Pose poseRequestObtained() {
		System.out.println("Start: " + model.getRobot().getPose());
		return model.getRobot().getPose();
	}

	// Called from PCCommunicator when the robot sends a pose (e.g., estimated pose)
	public void poseObtained(int traceID, double x, double y, double angle) {
		Pose p = new Pose(0, x, y, angle);
System.out.println("Trace ID: " + traceID + "   " + p);
		model.getRobot().setPose(p);
		model.getTraceWithID(traceID).append(model.getRobot().getPose());
	}

	// Called from PCCommunicator when the robot sends an obstacle estimate
	public void obstacleObtained(int index, double x, double y) {
		Trace t = model.getTraces().get(0);
		Position p = new Position(x, y);
System.out.println("Obstacle (Set: " + (t.getObstacles().get(index).getDescription()) + ")   " + p);
		t.getObstacleSet(index).add(p);
	}

	// Called from PCCommunicator when the robot sends a sensor reading
	public void readingObtained(int dataSetIndex, float x, float y, float angle, float reading, float angleOffset, float dError, float beamAngle) {
		Trace t = model.getTraces().get(0); /// Assume data is stored in the one and only trace
		model.getMap().applyModel(x*Environment.PIXELS_PER_CM(), y*Environment.PIXELS_PER_CM(), angle,
								  reading*Environment.PIXELS_PER_CM(), angleOffset, dError, beamAngle);
System.out.println("Reading (Set: " + (t.getObstacles().get(dataSetIndex).getDescription()) + ")   " + reading);
	}

	// Called from PCCommunicator when the robot sends a distance reading for localization purposes
	public void distanceObtained(float pixels) {
		model.getMap().getLocalizer().estimateFromReading(pixels);
model.getRobot().setPose(model.getMap().getLocalizer().guessPose());
		update();
	}

	// Called from PCCommunicator when the robot moves forward for localization purposes
	public void moveForwardObtained(float cm) {
		model.getMap().getLocalizer().updateLocation(cm*Environment.PIXELS_PER_CM());
model.getRobot().setPose(model.getMap().getLocalizer().guessPose());
		update();
	}

	// Called from PCCommunicator when the robot spins for localization purposes
	public void turnObtained(float degrees) {
		model.getMap().getLocalizer().updateOrientation(degrees);
model.getRobot().setPose(model.getMap().getLocalizer().guessPose());
		update();
	}

	// Called from PCCommunicator when the robot wants to reset a trace
	public void resetTrace(int traceID, double x, double y, double angle) {
		Pose p = new Pose(0, x, y, angle);
		model.getRobot().setPose(p);
		model.getTraceWithID(traceID).reset(x, y, angle);
		model.getMap().clear();
	}

	// Called from PCCommunicator when the robot wants to reset a map
	public void resetMap() {
		model.getMap().clear();
	}

	// Called from PCCommunicator when an incoming debug string arrives
	public void debugPrint(String s) {
		debugDialog.setVisible(true);
		debugDialog.appendTextWithCR(s);
	}

	public void update() {
		int xs, ys, xe, ye;
		xs = (int)model.getRobot().getPose().x * Environment.PIXELS_PER_CM();
		ys = (int)model.getRobot().getPose().y * Environment.PIXELS_PER_CM();
		model.getMap().getPathPlanner().setStart(new Point(xs,ys));
		xe = (int)model.getDestination().x * Environment.PIXELS_PER_CM();
		ye = (int)model.getDestination().y * Environment.PIXELS_PER_CM();
		model.getMap().getPathPlanner().setEnd(new Point(xe,ye));

		// Recompute the visibility graph if needed
		if (view.getShowVisibilityButton().isSelected() || view.getShowPathButton().isSelected() ||
			view.getShowTurnPathButton().isSelected() || view.getShowGrownButton().isSelected()) {
			if (view.getShowGrownButton().isSelected()) {
				ArrayList<Polygon> triangles = model.getMap().getPathPlanner().splitIntoTriangles(model.getMap().getPolygons());
				ArrayList<Polygon> polys = model.getMap().getPathPlanner().computeGrownObstacleSpace(triangles, model.getRobot().getDiameter()/2*Environment.PIXELS_PER_CM(), OccupancyGrid.ROUNDNESS);
				polys = model.getMap().createFrom(model.getMap().getPolygons(), polys, model.getMapImage());  // Create a map since we will be tracing the polygons
				model.getMap().getGrownPolygons().clear();
				model.getMap().getGrownPolygons().addAll(polys);
				model.getMap().computeVisibilityGraph(model.getMap().getPathPlanner().splitIntoTriangles(model.getMap().getGrownPolygons()));
			}
			else {
				ArrayList<Polygon> triangles = model.getMap().getPathPlanner().splitIntoTriangles(model.getMap().getPolygons());
				model.getMap().getPolygons().clear();
				model.getMap().getPolygons().addAll(triangles);
				model.getMap().computeVisibilityGraph(model.getMap().getPolygons());
			}

			if (view.getShowPathButton().isSelected() || view.getShowTurnPathButton().isSelected()) {
				Graph g = model.getMap().getGraph();
				if ((g != null) && (g.node(xs,ys) != null) && (g.node(xe,ye) != null)) {
					if (view.getShowPathButton().isSelected())
						model.getMap().getPathPlanner().computeShortestPath(g, g.node(xs,ys), g.node(xe,ye));
					else
						model.getMap().getPathPlanner().computeShortestTurnPath(g, g.node(xs,ys), g.node(xe,ye));
				}
			}
		}

		view.update();
	}

	public static void main(String[] args) {
		JFrame frame = new RTMS();
        frame.setVisible(true);
    }
}