import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.geom.*;
import javax.swing.JOptionPane;

// This class represents a trace of a robot's movements as a sequence of consecutive poses
public class Trace {
	private static int 	LATEST_ID = 0;

	private int						ID;				// unique ID of the trace
	private String					name;			// label of the trace
	private Color					color;			// color to draw the trace
	private ArrayList<Pose>			poses;			// poses of the robot over time
	private ArrayList<ObstacleSet>	obstacles;		// set of obstacle estimates obtained from various sensors over time
	private boolean					visible;		// true of poses are to be shown as circles
	private boolean					showConnected;	// true if poses should be connected together
	private boolean					showRobotOutline; // true if robot outline should be shown

	public Trace(String n) {
		this(n, Color.RED);
	}
	public Trace(String n, Color c) {
		ID = ++LATEST_ID;
		name = n;
		color = c;
		poses = new ArrayList<Pose>();
		obstacles = new ArrayList<ObstacleSet>();
		visible = true;
		showConnected = true;
		showRobotOutline = false;
	}

	public void reset(double x, double y, double angle) {
		poses.clear();
		poses.add(new Pose(0, x, y, angle));
		for (ObstacleSet ob: obstacles) {
			ob.reset();
			ob.add(new Position(x, y));
		}
	}

	// This is what will be displayed in the Traces list e.g., "Grid Estimate (ID:34)"
	public String toString() { return name + " (ID:" + ID + ")"; }

	// Get/Set methods
	public int getID() { return ID; }
	public String getName() { return name; }
	public Color getColor() { return color; }
	public boolean isVisible() { return visible; }
	public boolean isConnected() { return showConnected; }
	public boolean isOutlined() { return showRobotOutline; }
	public ArrayList<Pose> getPoses() { return poses; }
	public ArrayList<ObstacleSet> getObstacles() { return obstacles; }

	// Return the obstacle set with the given name
	public ObstacleSet getObstacleSet(int index) {
		if (obstacles.isEmpty()) return null;
		return obstacles.get(index);
	}

	public void setColor(Color c) { color = c; }
	public void setVisible(boolean b) { visible = b; }
	public void setShowConnected(boolean b) { showConnected = b; }
	public void setShowRobotOutline(boolean b) { showRobotOutline = b; }

	public void append(Pose p) {
		poses.add(p);
	}

    public void draw(Graphics g, Color c, Environment e, Robot r) {
    	if (showRobotOutline) {
    		for (Pose p: poses)
    			r.drawWithPose(g, p, e, color.darker());
    	}
	   	if (showConnected) {
    		g.setColor(color.darker());
			for (int i=0; i<poses.size()-1; i++) {
				Pose p1 = poses.get(i);
				Pose p2 = poses.get(i+1);
				g.drawLine((int)(p1.x*Environment.PIXELS_PER_CM()), e.getHeightInPixels() - (int)(p1.y*Environment.PIXELS_PER_CM()),
						   (int)(p2.x*Environment.PIXELS_PER_CM()), e.getHeightInPixels() - (int)(p2.y*Environment.PIXELS_PER_CM()));
			}
    	}
		if (visible) {
	    	for (int i=0; i<poses.size(); i++) {
				g.setColor(color);
	    		Pose p = poses.get(i);
				g.drawLine((int)(p.x*Environment.PIXELS_PER_CM()), e.getHeightInPixels() - (int)(p.y*Environment.PIXELS_PER_CM()),
						   (int)(p.x*Environment.PIXELS_PER_CM() + 5*Math.cos(Math.toRadians(p.angle))),
						   e.getHeightInPixels() - (int)(p.y*Environment.PIXELS_PER_CM() + 5*Math.sin(Math.toRadians(p.angle))));
				g.fillOval((int)(p.x*Environment.PIXELS_PER_CM())-2, e.getHeightInPixels() - (int)(p.y*Environment.PIXELS_PER_CM())-2, 4, 4);
				g.setColor(java.awt.Color.black);
				g.drawOval((int)(p.x*Environment.PIXELS_PER_CM())-2, e.getHeightInPixels() - (int)(p.y*Environment.PIXELS_PER_CM())-2, 4, 4);
			}
    	}
    	if ((showConnected || visible) && !poses.isEmpty())
    		r.drawWithPose(g, poses.get(poses.size()-1), e, color.darker());
    }

    public void saveTo(String fileName) {
    	try {
    		PrintWriter file = new PrintWriter(new FileWriter(fileName));
    		file.println("x,y,angle,sonar,ir");
    		for (int i=1; i<poses.size(); i++) {
    			Pose p = poses.get(i);
    			file.print(String.format("%d,%d,%d",(int)p.x,(int)p.y,(int)p.angle));
    			for (ObstacleSet os: obstacles) {
    				file.print(String.format(",%d", (int)Point2D.distance(os.getObstacle(i-1).x, os.getObstacle(i-1).y, p.x, p.y)));
    			}
    			file.println();
    		}
    		file.close();
    	}
    	catch(IOException ex) {
			JOptionPane.showMessageDialog(null, "Cannot save data to trace file: " + fileName, "Error", JOptionPane.ERROR_MESSAGE);
    	}

    }
}